/*******************************************************************************
	xfer9860 - a Casio fx-9860G (SD) communication utility
	Copyright (C) 2007
		Manuel Naranjo <naranjo.manuel@gmail.com>
		Andreas Bertheussen <andreasmarcel@gmail.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	MA  02110-1301, USA.
*******************************************************************************/

#include <stdio.h>
#include "libusb/libusb.h"
#include <string.h>

#include "Casio9860.h"
#include "usbio.h"
#include "config.h"

static void print_devs(libusb_device **dev); //prototype of the function

//Check For Our VID & PID
int is_usbdevblock( libusb_device *dev )
{
	struct libusb_device_descriptor desc;
	libusb_get_device_descriptor( dev, &desc );

	if( desc.idVendor == C9860_VENDOR_ID && desc.idProduct == C9860_PRODUCT_ID ){
		return 1;
	}

	return 0;
}

libusb_device **devs; //pointer to pointer of device, used to retrieve a list of devices
libusb_context *ctx = NULL; //a libusb session

struct libusb_device_handle* InitDevice() {
    libusb_device_handle *dev_handle = NULL; //a device handle
    ssize_t cnt; //holding number of devices in list
    //initialize a library session
    if(libusb_init(&ctx) < 0) {
        printf("\n[E] ʼ\n") ;
        return NULL;
    }
    else printf("\n[I] ʼɹ\n") ;

    libusb_set_debug(ctx, 3); //set verbosity level to 3, as suggested in the documentation
    cnt = libusb_get_device_list(ctx, &devs); //get the list of devices
    if(cnt < 0) {
        printf("\n[E] ȡ豸бʧܡ\n") ;
        return NULL;
    }
    else printf("\n[I] ȡ豸бɹ\n") ;



    //print_devs(devs); //print specs of this device


    ssize_t i = 0;
    struct libusb_device *found = NULL;

	// find our device
	for(i = 0; i < cnt; i++){
		libusb_device *device = devs[i];
		if( is_usbdevblock(device) ){
			found = device;
			break;
		}
	}
    if(found){
        printf("\nҵ\n") ;
        if (libusb_open(found, &dev_handle)){
			printf("Unable to open usb device\n");
		}
    }

    //USB\VID_07CF&PID_6101&REV_0100

    //dev_handle = libusb_open_device_with_vid_pid(ctx, 1999, 24833); //these are vendorID and productID I found for my usb device
    if(dev_handle == NULL){
        printf("\n[E] 򿪼豸ʧܡ\n") ;
        return NULL ;
    }
    else printf("\n[I] 򿪼豸ɹ\n") ;
    return dev_handle;
}


//static void print_devs(libusb_device **devs)
//{
//	libusb_device *dev;
//	int i = 0;
//
//	while ((dev = devs[i++]) != NULL) {
//		struct libusb_device_descriptor desc;
//		if (libusb_get_device_descriptor(dev, &desc) < 0) {
//			fprintf(stderr, "failed to get device descriptor");
//			return;
//		}
//        printf("Number of possible configurations: %d\n",(int)desc.bNumConfigurations);
//        printf("Device Class: %d\n",(int)desc.bDeviceClass);
//        printf("VendorID: %d\n",desc.idVendor);
//        printf("ProductID: %d\n",desc.idProduct);
//
//		//printf("%04x:%04x (bus %d, device %d)\n",
////			desc.idVendor, desc.idProduct,
////			libusb_get_bus_number(dev), libusb_get_device_address(dev));
//	}
//
//
//}

//void printdev(libusb_device *dev) {
//    struct libusb_device_descriptor desc;
//    struct libusb_config_descriptor config;
//    const struct libusb_interface *inter;
//    const struct libusb_interface_descriptor *interdesc;
//    const struct libusb_endpoint_descriptor *epdesc;
//
//    int r = libusb_get_device_descriptor(dev, &desc);
//    if (r < 0) {
//        printf("failed to get device descriptor");
//        return;
//    }
//
//    printf("Number of possible configurations: %d\n",(int)desc.bNumConfigurations);
//    printf("Device Class: %d\n",(int)desc.bDeviceClass);
//    printf("VendorID: %d\n",desc.idVendor);
//    printf("ProductID: %d\n",desc.idProduct);
//
//    r = libusb_get_config_descriptor(dev, 0, &config);
//    if(r==0)
//        printf("Interfaces: %d\n",(int)config.bNumInterfaces);
//    else printf("ȡʧ\n") ;
//
//
//
////    int i ;
////    for(i=0; i<(int)config.bNumInterfaces; i++) {
////        inter = &(config.interface[i]);
////
////        printf("Number of alternate settings: %d",inter->num_altsetting);
////        int j ;
////        for(j=0; j<inter->num_altsetting; j++) {
////            interdesc = &inter->altsetting[j];
////            printf("Interface Number: %d",(int)interdesc->bInterfaceNumber);
////            printf("Number of endpoints: %d",(int)interdesc->bNumEndpoints);
////
////            int k;
////            for(k=0; k<(int)interdesc->bNumEndpoints; k++) {
////                epdesc = &interdesc->endpoint[k];
////                printf("Descriptor Type: %d",(int)epdesc->bDescriptorType);
////                printf("EP Address: %d",(int)epdesc->bEndpointAddress);
////            }
////        }
////    }
////    libusb_free_config_descriptor(&config);
//}


//
//struct libusb_device_handle* fx_getDeviceHandle() {
//	int ret = 0;
//	struct usb_device *usb_dev;
//	struct libusb_device_handle *usb_handle;
//
//	usb_dev = (struct usb_device*)device_init();
//	if (usb_dev == NULL) {
//		return NULL; /* this is a common 'error', won't print error message */
//	}
//
//	usb_handle = (struct libusb_device_handle*)usb_open(usb_dev);
//	if (usb_handle == NULL) {
//		printf("\nERR: usb_open() returned NULL.\n");
//		return NULL;
//	}
//
//	ret = usb_set_configuration(usb_handle, 1);
//	if (ret < 0) { /* needed on WIN32 */
//		printf("\nERR: usb_set_configuration(): %i\n", ret);
//		usb_close(usb_handle);
//		return NULL;
//	}
//
//	ret = usb_claim_interface(usb_handle, 0);
//	if (ret < 0) {
//		printf("\nERR: usb_claim_interface(): %i\n", ret);
//		usb_close(usb_handle);
//		return NULL;
//	}

    /* bMaxPacketSize */
//	printf("\ndevnum:%d\tnum_children%d\n",usb_dev->devnum,usb_dev->num_children) ;
//    printf("des.bMaxPacketSize0:%d\tconfig.wMaxPacketSize:%d\n",usb_dev->descriptor.bMaxPacketSize0,usb_dev->config->interface->altsetting->endpoint->wMaxPacketSize) ;
//
//    printf("bNumEndpoints:%d\n",usb_dev->config->interface->altsetting->bNumEndpoints) ;
//    printf("Endpoint_address:%d\n",usb_dev->config->interface->altsetting->endpoint[0].bEndpointAddress) ;
//    printf("Endpoint_address:%d\n",usb_dev->config->interface->altsetting->endpoint[1].bEndpointAddress) ;
//    printf("Endpoint_address:%d\n",usb_dev->config->interface->altsetting->endpoint[2].bEndpointAddress) ;
//    printf("bSynchAddress:%d\n",usb_dev->config->interface->altsetting->endpoint->bSynchAddress) ;
//
//
//	return usb_handle;
//}

//void fx_releaseDeviceHandle(struct libusb_device_handle* usb_handle) {
//	usb_release_interface(usb_handle, 0);
//	return;
//}

//struct usb_device *device_init(void) {
//	struct usb_bus *usb_bus;
//	struct usb_device *dev;
//	usb_init();
//	usb_find_busses();
//	usb_find_devices();
//	for (usb_bus = usb_busses; usb_bus; usb_bus = usb_bus->next) {
//		for (dev = usb_bus->devices; dev; dev = dev->next) {
//			if ((dev->descriptor.idVendor == C9860_VENDOR_ID) && (dev->descriptor.idProduct == C9860_PRODUCT_ID))
//				return dev;
//		}
//	}
//	return NULL;
//}
int LIBUSB_CALL libusb_control_transfer(libusb_device_handle *dev_handle,
	uint8_t request_type, uint8_t bRequest, uint16_t wValue, uint16_t wIndex,
	unsigned char *data, uint16_t wLength, unsigned int timeout);

int fx_initDevice(struct libusb_device_handle *usb_handle) {
	int ret;
	char* buffer;
	buffer = (char*)calloc(0x29, sizeof(char));

	ret = libusb_control_transfer(usb_handle, 0x80, 0x6, 0x100, 0, buffer, 0x12, 200);
	debug(1, buffer, ret);

	if (ret < 0) {
		fprintf(stderr, "\nERR: fx_initDevice(), 1'st control: %i \n", ret);
		goto exit;
	}

	ret = libusb_control_transfer(usb_handle, 0x80, 0x6, 0x200, 0, buffer, 0x29, 250);
	debug(1, buffer, ret);
	if (ret < 0) {
		fprintf(stderr, "\nERR: fx_initDevice(), 2'nd control: %i \n", ret);
		goto exit;
	}

	ret = libusb_control_transfer(usb_handle, 0x41, 0x1, 0x0, 0, buffer, 0x0, 250);
	debug(1, buffer, ret);
	if (ret < 0) {
		fprintf(stderr, "\nERR: fx_initDevice(), 3'rd control: %i \n", ret);
		goto exit;
	}
	ret = 0;
exit:
	free(buffer);
	return ret;
}

int Connect_Complete() {
    int BytesWritten = 0;
    int ret ;
    unsigned char OutputPacketBuffer[65];

    libusb_clear_halt(usb_handle,0x82);
    libusb_clear_halt(usb_handle,0x1);

    ret = libusb_bulk_transfer(usb_handle,0x1,"\x01\x30\x30\x30",4, &BytesWritten,USB_WRITE_TIMEOUT);
    ret = libusb_bulk_transfer(usb_handle,0x1,"\x37\x30",2, &BytesWritten,USB_WRITE_TIMEOUT);

    if (ret >= 0) {
    	printf("\n[>] ر.. ");
    	printf("!\n");
    }
    libusb_close(usb_handle);
    libusb_free_device_list(devs, 1); //free the list, unref the devices in it
    libusb_exit(ctx); //close the session
}


int fx_sendVerify(struct libusb_device_handle *usb_handle, char *buffer, char *type) {

	/* Type: 0x05
	 * ST: 00 or 01*/
	memcpy(buffer, "\x05\x30\x30\x30", 4);
	if (type[1] == '1') {
		memcpy(buffer+2, "\x31", 1);
	}
	fx_appendChecksum(buffer, 4);
	return WriteUSB(usb_handle, buffer, 6);
}

int fx_doConnVer(struct libusb_device_handle *usb_handle) {
    int ret ;
	char *buffer = (char*)calloc(6, sizeof(char));
	if (buffer == NULL) {printf("ERR: fx_doConnVer(): allocation failed.\n");}

    ReadCNUL(usb_handle, buffer) ;
    fx_sendVerify(usb_handle, buffer, "00");	/* sends connver for start of communication */

	ReadUSB(usb_handle, buffer, 6);
	ret = fx_getPacketType(buffer) ;

//    printf("\n!%d!\n",ret) ;
//    if (ret != T_POSITIVE && ret != 0x5){
////        ReadUSB(usb_handle, buffer, 6);
////        ret = fx_getPacketType(buffer) ;
////      printf("\n!%d!\n",ret) ;
//	}

	if (ret == T_POSITIVE) {/* assuming the positive response is 'plain' */
		free(buffer);
		return 0;
	} else {
		free(buffer);
		return 1;	/* failure */
	}
}

//int fx_sendTerminate(struct libusb_device_handle *usb_handle, char *buffer) {
//	/* Type: 0x18
//	 * ST: 01 */
//	memcpy(buffer, "\x18\x30\x31\x30\x36\x46", 6);
//	return WriteUSB(usb_handle, buffer, 6);
//}
//
int fx_sendPositive(struct libusb_device_handle *usb_handle, char *buffer, char type) {
	/* Type: 0x06
	 * ST: given as argument */
	memcpy(buffer, "\x06\x30\x30\x30", 4);
	if (type == POSITIVE_OVERWRITE || type == POSITIVE_SYSINFO) {
		memcpy(buffer+2, &type, 1);
	}

	fx_appendChecksum(buffer, 4);
	return WriteUSB(usb_handle, buffer, 6);
}

int fx_getPacketType(char *buffer) {
	if (buffer != NULL && buffer[0] != 0x00)
		return buffer[0];
	else
		return -1;
}

int fx_sendNegative(struct libusb_device_handle *usb_handle, char *buffer, char type) {
	/* Type 0x05
	 * ST: given as argument */
	memcpy(buffer, "\x05\x30\x30\x30", 4);
	if (type >= 0x30 || type <= 0x36) {	/* from '0' and '6' */
		memcpy(buffer+2, &type, 1);
	}
	fx_appendChecksum(buffer, 4);
	return WriteUSB(usb_handle, buffer, 6);
}

int fx_sendChange_Direction(struct libusb_device_handle *usb_handle, char *buffer) {
	memcpy(buffer, "\x03\x30\x30\x30\x37\x30", 6);
	return WriteUSB(usb_handle, buffer, 6);
}

int fx_sendFlash_Capacity_Request(struct libusb_device_handle *usb_handle, char *buffer, char *device) {
	short int len = strlen(device);
	memcpy(buffer, "\x01\x34\x42\x31", 4);
	sprintf(buffer+OFF_DS, "%04X", 24+len);
	memcpy(buffer+OFF_OW, "000000000000", 12); /* includes OW, DT and FS */
	sprintf(buffer+OFF_SD1, "00000000%02X00", len);	/* includes D1 - D6 */
	memcpy(buffer+OFF_D1, device, len);

	fx_appendChecksum(buffer, 32+len);

	return WriteUSB(usb_handle, buffer, 34+len);
}

int fx_sendMCSCapacityRequest(struct libusb_device_handle *usb_handle, char *buffer) {
	memcpy(buffer, "\x01\x32\x42\x30", 4);	/* ST is 42 */
	fx_appendChecksum(buffer, 4);

	return WriteUSB(usb_handle, buffer, 6);
}

int fx_sendFlashCollectGarbage(struct libusb_device_handle *usb_handle, char *buffer, char *device) {
	short int len = strlen(device);
	memcpy(buffer, "\x01\x35\x31\x31", 4);	/* ST 51 */
	sprintf(buffer+OFF_DS, "%04X", 24+len);
	memcpy(buffer+OFF_OW, "000000000000", 12);
	sprintf(buffer+OFF_SD1, "00000000%02X00", len);
	memcpy(buffer+OFF_D1, device, len);

	fx_appendChecksum(buffer, 32+len);

	return WriteUSB(usb_handle, buffer, 34+len);

}

int fx_getFlashCapacity(struct libusb_device_handle *usb_handle, char *device) {
	int freeSize = 0;
	char * buffer = (char*)calloc(40,sizeof(char));
		if (buffer == NULL) { printf("ERR: fx_getFlashCapacity: alloc error\n"); return -1; }

	fx_sendFlash_Capacity_Request(usb_handle, buffer, device);
	ReadUSB(usb_handle, buffer, 6);
		if (fx_getPacketType(buffer) != T_POSITIVE) { printf("ERR: fx_getFlashCapacity: no proper response\n"); return -1; }

	fx_sendChange_Direction(usb_handle, buffer);
	ReadUSB(usb_handle, buffer, 0x26);
		if (fx_getPacketType(buffer) != T_COMMAND) { printf("ERR: fx_getFlashCapacity: no returned command\n"); return -1; }

	freeSize = fx_asciiHexToInt(buffer+OFF_FS, 8); /* free size is returned in FS */

	fx_sendPositive(usb_handle, buffer, POSITIVE_NORMAL);
	ReadUSB(usb_handle, buffer, 6);

	free(buffer);

	return freeSize; /* converts 'hex-ascii' to int */
}

int fx_getMCSCapacity(struct libusb_device_handle *usb_handle) {
	int freeSize = 0;
	char * buffer = (char*)calloc(40,sizeof(char));
	fx_sendMCSCapacityRequest(usb_handle, buffer);
	ReadUSB(usb_handle, buffer, 6);
		if (fx_getPacketType(buffer) != T_POSITIVE) { printf("ERR: fx_getMCShCapacity: no proper response\n"); return -1; }
	fx_sendChange_Direction(usb_handle, buffer);

	ReadUSB(usb_handle, buffer, 0x26);
		if (fx_getPacketType(buffer) != T_COMMAND) { printf("ERR: fx_getMCSCapacity: no returned command\n"); return -1; }

	freeSize = fx_asciiHexToInt(buffer+OFF_FS, 8);

	fx_sendPositive(usb_handle, buffer, POSITIVE_NORMAL);
	ReadUSB(usb_handle, buffer, 6);

	free(buffer);

	return freeSize;
}

int fx_sendFlashFileTransmission(struct libusb_device_handle *usb_handle, char *buffer, int filesize, char *filename, char *device) {
	short int fnsize = strlen(filename);
	short int devsize = strlen(device);
	memcpy(buffer, "\x01\x34\x35\x31", 4); /* T, ST = 45, DF */
	sprintf(buffer+OFF_DS, "%04X", 24+fnsize+devsize);
	memcpy(buffer+OFF_OW, "0280", 4); /* OW, DT */
	sprintf(buffer+OFF_FS, "%08X", filesize); /* FS   8 byte*/
	sprintf(buffer+OFF_SD1, "00%02X0000%02X00", fnsize, devsize); /* SD1 - SD6, 12b*/
	sprintf(buffer+OFF_D1, "%s%s", filename, device);

	fx_appendChecksum(buffer, 32+fnsize+devsize);

	return WriteUSB(usb_handle, buffer, 34+fnsize+devsize);
}

int fx_sendFlashFileTransmissionRequest(struct libusb_device_handle *usb_handle, char* buffer, char* filename, char* device) {
	short int fnsize = strlen(filename);
	short int devsize = strlen(device);
	memcpy(buffer, "\x01\x34\x34\x31", 4); /* T, ST = 44, DF */
	sprintf(buffer+OFF_DS, "%04X", 24+fnsize+devsize);
	memcpy(buffer+OFF_OW, "0000", 4); /* OW, DT */
	sprintf(buffer+OFF_FS, "00000000"); /* FS   8 byte*/
	sprintf(buffer+OFF_SD1, "00%02X0000%02X00", fnsize, devsize); /* SD2 and SD5 are in use, no folder support yet */
	sprintf(buffer+OFF_D1, "%s%s", filename, device);

	fx_appendChecksum(buffer, 32+fnsize+devsize);

	return WriteUSB(usb_handle, buffer, 34+fnsize+devsize);
}

int fx_sendData(struct libusb_device_handle *usb_handle, char *buffer, char* subtype, int total, int number, char *data, int length) {
	memcpy(buffer, "\x02\x00\x00\x31", 4); /* T, DF, leaves a hole for ST */
	 memcpy(buffer+1, subtype, 2); /* ST */
	sprintf(buffer+OFF_DS, "%04X", 8+length); /* DS, 4 b */
	sprintf(buffer+OFF_TP, "%04X%04X", total, number);
	memcpy(buffer+OFF_DD, data, length);

	fx_appendChecksum(buffer, 16+length);
	return WriteUSB(usb_handle, buffer, 18+length);
}

int fx_appendChecksum(char *buffer, int length) {
	int i;
	char sum = 0;
	for (i = 1; i < length; i++) {
		sum += *(buffer+i);
	}
	sprintf(buffer+length, "%02X", ((~sum)+1) & 0xFF);
	/* The value appears sometimes as FFFFxx, where AND'ing it with 0xFF, you
	 * get the wanted one-byte 0000xx.. */
	return 0;
}

int fx_unescapeBytes(char *source, char*dest, int length) {
	int i = 0, j = 0;
	while (i < length) {
		if (source[i] == 0x5C) { /* if we find an escape character.. look at next byte */
			if (source[i+1] == 0x5C) { /* if it's another escape char, it is taken literally */
				dest[j] = source[i];
				i++;	/* and we skip the following byte */
				goto done;
			} else {
				dest[j] = source[i+1]-0x20; /* if its a usual byte, subtract 0x20 */
				i++;
				goto done;
			}
		}

		/* default copying */
		dest[j] = source[i];

	done:
		i++;
		j++;
	}

	return j;
}

int fx_escapeBytes(char *source, char *dest, int length) {
	int i = 0, j = 0;
	while(i < length) {
		if (source[i] < 0x20) {
			dest[j] = 0x5C;
			dest[j+1] = 0x20+source[i];
			j++;
			goto done;
		}

		if (source[i] == 0x5C) {
			dest[j] = 0x5C; dest[j+1] = 0x5C; j++; /* I might have to use 0x7C here */
			goto done;
		}
		/* default action */
		*(dest+j) = *(source+i);

	done:
		j++;
		i++;
	}
	return j; /* length of destination (new) buffer */
}

long int fx_asciiHexToInt(char *source, int length) {
	int value;
	char *tmp = (char*)calloc(length+1, sizeof(char));
	memcpy(tmp, source, length); tmp[length] = 0; /* using the extra byte for null-termination */
	value = strtol(tmp, NULL, 16);
	free(tmp);
	return value;
}
